import glob
import os
import re
import openpyxl as ox
import pandas as pd
import xlsxwriter

def write_yaml():
    return {'FORMULATION DIRECTORY': None,
            'RESULT STRUCTURED': True,
            'RESULTS DIRECTORY': None,
            'TOPIC STRUCTURED': False,
            'PROJECT DIRECTORY': None,
            'OUTPUT DIRECTORY': None
    }

def data_start(sheet, header=False):
    y = sheet.values
    _header = []
    for x in y:
        if header:
            _header.append(x)
        if x[0] == '67 6f':
            return _header, y, True
        if x[0] == '6e 74':
            return _header, y, False
    return _header, y, False


def data_extraction(wb, sheet, use_header=False):
    dataframe = pd.DataFrame()
    header = pd.DataFrame()
    if sheet in wb.sheetnames:
        data_wb = wb[sheet]
        head_data, data, transpose = data_start(data_wb, use_header)
        if transpose:
            data_transposed = [transposed for transposed in zip(*[line for line in data if line[0]])]
            dataframe, header = _extraction(data_transposed, head_data, header, dataframe, use_header)
        else:
            data_ = [line for line in data]
            dataframe, header = _extraction(data_, head_data, header, dataframe, use_header)
            dataframe = dataframe[dataframe.columns.dropna()]
    return [dataframe, [header, dataframe]][use_header]


def _extraction(data, head_data, header, dataframe, use_header):
    if data:
        dataframe = pd.DataFrame(columns=data[0], data=[line for line in data[1:]])
        dataframe = dataframe.truncate(after=dataframe.last_valid_index())
    if use_header:
        head_data = [line for line in head_data]
        if head_data:
            header = pd.DataFrame(data=[line for line in head_data]).dropna(axis=1, how='all')
    return dataframe, header


def workbook_save(name, path, dataframe, sheet, **kwargs):
    data_path = os.path.join(path, name + '.xlsx')
    writer = pd.ExcelWriter(data_path, engine='xlsxwriter')
    dataframe.to_excel(writer, sheet_name=sheet, startrow=1, header=False, index=False)
    if kwargs.get('header'):
        workbook = writer.book
        header_format = workbook.add_format(kwargs.get('header'))
        for _sheet in writer.sheets:
            worksheet = writer.sheets[_sheet]
            for col_num, value in enumerate(dataframe.columns.values):
                worksheet.write(0, col_num, value, header_format)
    writer.save()


def workbook_load_file(path, write=False) -> [ox.workbook]:
    results_wb = []
    excels = results_excels(path)
    for e in excels:
        e_path = os.path.join(path, e)
        if write:
            wb = ox.load_workbook(filename=e_path, data_only=True)
        else:
            wb = ox.load_workbook(filename=e_path, data_only=True,
                                  read_only=True)
        results_wb.append(wb)
    return results_wb


def workbook_load_path(path, write=False):
    data_file = [x for x in os.listdir(path) if '.xlsx' in os.path.splitext(x)]
    if len(data_file) > 1:
        # TODO: Logging
        pass
    else:
        pass
    data_file = data_file[0]
    data_path = os.path.join(path, data_file)
    if write:
        wb = ox.load_workbook(filename=data_path, data_only=True)
    else:
        wb = ox.load_workbook(filename=data_path, data_only=True,
                              read_only=True)
    return wb


def rename_dictionary():
    return {"pg": "propylene glycol", 'propylen glycol': 'propylene glycol"', "glycerine": "glycerin"}


def topic_directories(path) -> (list, list):
    topics = [topic for topic in os.listdir(path) if re.search(r'^Topic', topic)]

    directories = [os.path.join(path, topic) for topic in topics if
                   os.path.isdir(os.path.join(path, topic))]

    return topics, directories


def is_dir(path):
    path_list = [globed for globed in glob.glob(r'{}'.format(path + '/*')) if
                 os.path.isdir(globed)]
    return path_list


def results_excels(path) -> list:
    # Re to filter out temporary hidden files and junk files
    results = [excel for excel in os.listdir(path) if
               os.path.splitext(excel)[-1] == '.xlsx' and not re.search(r'^(~\$)',
                                                                        os.path.splitext(excel)[0]) and not os.stat(
                   os.path.join(path, excel)).st_size <= 1]
    return results



from crdatamgt.helpers import data_extraction, workbook_load_path, rename_dictionary
import pandas as pd
import os
from toolz import interleave



def _load_formulas(path, write=False):
    wb = workbook_load_path(path, write)
    return wb


def formulation_read(wb, header=False):
    p_header, p_formulation = data_extraction(wb, 'Formulations', header)
    p_formulation.columns = map(str.lower, p_formulation.columns)
    p_formulation.set_index('formulation id', inplace=True)
    p_formulation.rename(columns=rename_dictionary(), inplace=True)
    p_formulation = p_formulation.apply(pd.to_numeric, errors='coerce')

    return [p_formulation, [p_header, p_formulation]][header]


def tests_read(wb, header=False):
    p_header, p_tests_data = data_extraction(wb, 'Tests', header)
    p_tests_data.columns = map(str.lower, p_tests_data.columns)
    p_tests_data.set_index('formulation id', inplace=True)
    return [p_tests_data, [p_header, p_tests_data]][header]


def formulation_load(path):
    wb = _load_formulas(path)
    formulas = formulation_read(wb)
    return formulas


def formulation_write(formulas, formulation_path):
    return _write_formulations_tests(formulas, formulation_path)


def _write_formulations_tests(new_data, formulation_path):
    """
    This entire bit of code will become difficult to maintain. The entire concept of formulation flexibility should be
    revisited

    :param new_data:
    :param hold_blank_data:
    :return:
    """
    if new_data.dropna(how='all').empty:
        return pd.DataFrame()
    new_data.columns = map(str.lower, new_data.columns)
    new_data.rename(columns=rename_dictionary(), inplace=True)

    data_path = os.path.join(formulation_path, 'Formulation table.xlsx')
    writer = pd.ExcelWriter(data_path, engine='openpyxl')
    writer.book = workbook_load_path(formulation_path, True)
    formulation_wb = writer.book

    sheet_formulation = 'Formulations'
    sheet_tests = 'Tests'

    formulation_header, old_formula_data = formulation_read(formulation_wb, True)
    test_header, test_data = tests_read(formulation_wb, True)

    new_formula_data = new_data.drop_duplicates().drop(columns=['formulation'], errors='ignore').dropna(how='all')
    formulation_only_test_data = new_data.loc[:, ['test', 'formulation']].dropna()

    test_set = old_formula_data.merge(new_formula_data, indicator=True, how='outer')
    test_set.index += 1
    test_set.drop(columns=['formulation', 'units'], errors='ignore', inplace=True)
    check = test_set.drop(columns=['test'], errors='ignore').sum(axis=1).between(100, 100)

    gf = test_set[check]
    bf = test_set[~check]

    # if not bf.empty:
    #     try:
    #         f_log.info(f"Bad Formulas present in {topic_path}")
    #     except PermissionError as e:
    #         f_log2 = log.get_logger(f'formulation {e}')
    #         f_log2.info(f"Bad Formulas present in {topic_path}")

    new_formulas = gf[gf['_merge'] == 'right_only'].drop(columns=['_merge'], errors='ignore').drop_duplicates()
    updated_data = old_formula_data.append([new_formulas])
    updated_data.reset_index(inplace=True)
    updated_data.rename(columns={"index": "Formulation ID"}, inplace=True)
    updated_data.index += 1

    new_tests = gf['test'].dropna()
    for formula in gf['test'].index:
        if not (test_data.index == formula).any():
            test_data.loc[formula] = ""
    for test in new_tests.index:
        m = gf['test'][test].lower()
        test_data.loc[test, m] = "x"
    for test in formulation_only_test_data.index:
        _formula = formulation_only_test_data.iloc[test].get('formulation')
        _test = formulation_only_test_data.iloc[test].get('test').lower()
        test_data.loc[_formula, _test] = 'x'

    test_data.reset_index(inplace=True)

    replace_sheet(formulation_wb, sheet_formulation, writer, formulation_header, updated_data)
    replace_sheet(formulation_wb, sheet_tests, writer, test_header, test_data)

    writer.save()
    writer.close()

    formulation_return = pd.DataFrame()
    id_only = new_data['formulation'].dropna()

    formula_only = updated_data.set_index(['Formulation ID']).drop(columns=['test'], errors='ignore')
    new_formula_only = new_formula_data.drop(columns=['test', 'units'], errors='ignore')
    with_formulation = formula_only.merge(new_formula_only, indicator=True, how='outer')
    with_formulation.index += 1
    with_formulation.reset_index(inplace=True)
    with_formulation.rename(columns={"index": "Formulation ID"}, inplace=True)
    with_formulation = with_formulation.loc[with_formulation['_merge'] == 'both', 'Formulation ID']

    if not (id_only.empty and with_formulation.empty):
        # Flattening out the array
        idx = [fx for f in [with_formulation.values, id_only.values] for fx in f]
        formulation_id = pd.DataFrame(columns=['Formulation ID'], data=idx).drop_duplicates()
        used_formulations = updated_data.loc[formulation_id['Formulation ID']].drop(columns=['test']).set_index(
            'Formulation ID')
        max = pd.DataFrame(columns=[f'{q} Max' for q in used_formulations.columns],
                           data=[used_formulations.max().values])
        min = pd.DataFrame(columns=[f'{q} Min' for q in used_formulations.columns],
                           data=[used_formulations.min().values])
        formulation_min_max = pd.concat([max, min], axis=1)[list(interleave([max, min]))]
        formulation_id = pd.DataFrame(columns=['Formulation ID'],
                                      data=[', '.join([str(int(x[0])) for x in formulation_id.values])])
        formulation_return = formulation_id.join(formulation_min_max)

    return formulation_return


def replace_sheet(wb, sheet, writer, header, data):
    idx = wb.sheetnames.index(sheet)
    wb.remove(wb.worksheets[idx])
    header.to_excel(writer, sheet, index=False, header=False)
    data.drop(columns=['test'], errors='ignore').to_excel(writer, sheet, startrow=header.shape[0], startcol=0,
                                                          index=False)



import pandas as pd
import crdatamgt.topic as topic
from crdatamgt.helpers import topic_directories, workbook_load_file, workbook_save


def project_load_from_workbooks(workbooks, **kwargs):
    result_wb = workbooks
    project_sheets = []
    formatting = {}
    if kwargs.get('OUTPUT DIRECTORY'):
        output_path = kwargs.get('OUTPUT DIRECTORY')
    else:
        import os
        output_path = os.path.split(kwargs.get('RESULTS DIRECTORY'))[0]
    for result in result_wb:
        result_data = topic.read_topic(result, kwargs.get('FORMULATION DIRECTORY'))
        test = pd.concat(result_data.values(), axis=1, sort=False)
        project_sheets.append(test)
    compiled = pd.concat(project_sheets[::-1], sort=False).set_index('Topic ID').sort_index().reset_index()

    formatting['header'] = {
        'bold': True,
        'text_wrap': False,
        'valign': 'top',
        'fg_color': '#D7E4BC',
        'border': 1,
        'font_size': 16}

    # Write the column headers with the defined format.

    workbook_save("Project Results", output_path, compiled, 'Results', **formatting)


def project_load(**kwargs):
    if kwargs.get("TOPIC STRUCTURED"):
        topic_name, topic_path = topic_directories(kwargs.get('PROJECT DIRECTORY'))
        workbooks = [topic.load_topic(t_path) for t_path in topic_path]
        project_load_from_workbooks(workbooks, **kwargs)
    if kwargs.get('RESULT STRUCTURED'):
        workbooks = workbook_load_file(kwargs.get('RESULTS DIRECTORY'))
        project_load_from_workbooks(workbooks, **kwargs)
    else:
        # TODO: logging of badly formatted YAML
        pass



import crdatamgt
import yaml
import os

def main():
    def represent_none(self, _):
        return self.represent_scalar('tag:yaml.org,2002:null', '')

    yaml.add_representer(type(None), represent_none)

    try:
        with open("parameters.yaml", 'r') as stream:
            data_loaded = yaml.safe_load(stream)
            try:
                crdatamgt.project.project_load(**data_loaded)
            except TypeError:
                raise
                print("Please double-check your parameters.yaml document'")
                os.startfile(os.path.join(os.getcwd(), 'parameters.yaml'))
    except FileNotFoundError:
        data_loaded = crdatamgt.helpers.write_yaml()
        with open('parameters.yaml', 'w') as outfile:
            yaml.dump(data_loaded, outfile, default_flow_style=False)
        print("Please fill out the parameter.yaml document")
        os.startfile(os.path.join(os.getcwd(), 'parameters.yaml'))


if __name__ == "__main__":
    main()



from collections import defaultdict
from crdatamgt.formulations import formulation_write
from crdatamgt.helpers import workbook_load_path, data_extraction, rename_dictionary
import pandas as pd

def load_topic(path):
    print(path)
    wb = workbook_load_path(path)
    return wb


def read_topic(wb, formulation_path):
    df_dictionary = defaultdict()
    for sheet in wb.sheetnames:
        df_dictionary[sheet] = data_extraction(wb, sheet)
    if 'Formulation' in df_dictionary.keys():
        df_dictionary['Test'] = df_dictionary['Formulation']['Test'].drop_duplicates().dropna()
        if df_dictionary['Test'].isin(['Henry']).any():
            df_dictionary['Formulation Group'] = formulation_group(df_dictionary['Formulation'])
        df_dictionary['Formulation'] = formulation_write(df_dictionary.get('Formulation'), formulation_path)

    if 'Results' in df_dictionary.keys():
        if 'Replicate' in df_dictionary['Results']:
            df_dictionary['Results'] = pd.DataFrame(
                df_dictionary['Results'].set_index('Replicate').mean().round(2)).transpose().add_suffix(' Average')

    return df_dictionary


def formulation_group(formulas):
    temp_formulation_group = []
    result = formulas.drop(columns='Formulation').mean().round(0)
    fc = pd.DataFrame(result[result > 5]).transpose()
    fc.columns = map(str.lower, fc.columns)
    fc.rename(columns=rename_dictionary(), inplace=True)
    for item in fc:
        temp_formulation_group.append(f"{item} {fc[item].values[0]}")
    fg = ' '.join(temp_formulation_group)
    return pd.DataFrame(columns=['Formulation Group'], data=[fg])


def update_formulations(formulation_frame):
    return formulation_write(formulation_frame)
